import request from "supertest";
import { server } from "@/server";

import { Email } from "@/email/email";

import { NotFoundError } from "@snip/database/errors";
import { CollaborationInviteData } from "@snip/database/types";

const config = {
    send_email: false,
    allowedToInvite: false,
    isAlreadyCollaborator: false,
    invitedUserFound: false,
    inviteAlreadySent: false,
};

const resetConfig = () => {
    config.send_email = false;
    config.allowedToInvite = false;
    config.isAlreadyCollaborator = false;
    config.invitedUserFound = false;
    config.inviteAlreadySent = false;
};

vi.mock("../../email/send", async (importActual) => {
    const actual = await importActual<typeof import("../../email/send")>();
    return {
        sendEmail: async (email: Email) => {
            if (config.send_email) {
                return await actual.sendEmail(email);
            }
            return;
        },
    };
});

vi.mock("@snip/database", async (importActual) => {
    const actual = await importActual<typeof import("@snip/database")>();
    return {
        default: {
            ...actual,
            permission: {
                getByUserAndBookReduced: async (
                    user_id: number,
                    book_id: number,
                ) => {
                    return {
                        pACL: config.allowedToInvite,
                    };
                },
                getByEntityAndResource: async (
                    entity_id: number,
                    entity_type: string,
                    resource_id: number,
                    resource_type: string,
                ) => {
                    if (config.isAlreadyCollaborator) {
                        return {
                            pRead: true,
                            pWrite: true,
                            pDelete: true,
                            pACL: true,
                        };
                    }
                    throw new NotFoundError("permission", entity_id);
                },
            },
            user: {
                getByEmail: async (email: string) => {
                    if (config.invitedUserFound) {
                        return {
                            id: 1,
                            email,
                        };
                    }
                    throw new NotFoundError("user", email);
                },
                getById: async (user_id: number) => {
                    return {
                        id: user_id,
                        email: "test@test.de",
                    };
                },
            },
            book: {
                getById: async (book_id: number) => {
                    return {
                        id: book_id,
                        title: "test book title",
                    };
                },
            },
            collaboratorsInvite: {
                insert: async (
                    colab: CollaborationInviteData,
                ): Promise<Required<CollaborationInviteData>> => {
                    return {
                        book_id: colab.book_id,
                        user_id: colab.user_id,
                        invited_by: colab.invited_by,
                        expire_at: new Date(
                            Date.now() + 1000 * 60 * 60 * 24 * 7,
                        ),
                        accepted_at: null,
                        created_at: new Date(),
                        id: 1,
                    };
                },
                getPendingByBookIdAndUserId: async (
                    book_id: number,
                    user_id: number,
                ) => {
                    if (config.inviteAlreadySent) {
                        return [
                            {
                                book_id,
                                user_id,
                                invited_by: 1,
                                created_at: new Date(),
                            },
                        ];
                    }
                    return [];
                },
            },
        },
    };
});

describe("book router", () => {
    afterAll(() => {
        server.close();
    });

    beforeEach(() => {
        resetConfig();
    });

    describe("POST /email/book/inviteCollaborator", () => {
        describe("invalid requests", () => {
            it("400 if request body is invalid", async () => {
                const response = await request(server)
                    .post("/email/book/inviteCollaborator")
                    .send({ email: "test" });

                expect(response.status).toBe(400);
                expect(response.body).toEqual({
                    error: "Invalid request body",
                    details: "Valid Email required",
                });
            });

            it("400 if user is not allowed to invite", async () => {
                config.allowedToInvite = false;

                const response = await request(server)
                    .post("/email/book/inviteCollaborator")
                    .send({ email: "test@test.de", book_id: 1 });

                expect(response.status).toBe(400);
                expect(response.body).toEqual({
                    error: "Not allowed to invite to this book",
                });
            });

            it("400 if invited user not found", async () => {
                config.allowedToInvite = true;
                config.invitedUserFound = false;

                const response = await request(server)
                    .post("/email/book/inviteCollaborator")
                    .send({ email: "test@test.de", book_id: 1 });

                expect(response.status).toBe(400);
                expect(response.body).toEqual({
                    error: "User not found",
                });
            });

            it("400 if user is already collaborator", async () => {
                config.allowedToInvite = true;
                config.invitedUserFound = true;
                config.isAlreadyCollaborator = true;

                const response = await request(server)
                    .post("/email/book/inviteCollaborator")
                    .send({ email: "test@test.de", book_id: 1 });

                expect(response.status).toBe(400);
                expect(response.body).toEqual({
                    error: "Invited user is already a collaborator",
                });
            });

            it("400 if invite already sent", async () => {
                config.allowedToInvite = true;
                config.invitedUserFound = true;
                config.inviteAlreadySent = true;

                const response = await request(server)
                    .post("/email/book/inviteCollaborator")
                    .send({ email: "test@test.de", book_id: 1 });

                expect(response.status).toBe(400);
                expect(response.body).toEqual({
                    error: "Invite is still pending for this user you may try again later",
                    details:
                        "Invite is still pending for this user you may try again later",
                });
            });
        });

        it("200 if invite sent", async () => {
            config.allowedToInvite = true;
            config.invitedUserFound = true;

            const response = await request(server)
                .post("/email/book/inviteCollaborator")
                .send({ email: "test@test.de", book_id: 1 });

            expect(response.status).toBe(200);
            expect(response.body).toHaveProperty("success", true);
        });

        /** Uncomment to send emails as test!
         *
        describe("SEND EMAILS", () => {
            const addr = "sebastianbernd.mohr@uni-goettingen.de";
            it("collaborationInvite", async () => {
                config.allowedToInvite = true;
                config.invitedUserFound = true;
                config.send_email = true;
                
                const response = await request(server)
                    .post("/email/book/inviteCollaborator")
                    .send({
                        email: addr,
                        book_id: 1,
                    });

                expect(response.status).toBe(200);
                expect(response.body).toEqual({
                    success: true,
                    invite: {
                        book_id: 1,
                        user_id: 1,
                        invited_by: 1,
                    },
                });
            });
        });
        */
    });
});
