/** Account specific routers for the email service
 * in expressjs
 */
import express, { Router } from "express";
import { z } from "zod";

import { getPerms } from "@snip/auth/book_permissions";
import { signCollaborationRequestToken } from "@snip/auth/tokens/collaborationRequest";
import service from "@snip/database";
import { NotFoundError } from "@snip/database/errors";
import { ENTITY_USER, RESOURCE_BOOK } from "@snip/database/types";

import { authMiddleWare } from "@/common/auth";
import { CollaboratorRequestEmail } from "@/email/email";
import { sendEmail } from "@/email/send";

export const bookRouter: Router = express.Router();

bookRouter.post(
    "/inviteCollaborator",
    authMiddleWare,
    express.json(),
    async (req, res) => {
        const bodySchema = z.object({
            //User to invite
            email: z.string().email(),

            //Book id
            book_id: z.number().int(),
        });

        const parsedBody = bodySchema.safeParse(req.body);
        if (!parsedBody.success) {
            res.status(400).json({
                error: "Invalid request body",
                details: "Valid Email required",
            });
            return;
        }

        const { email, book_id } = parsedBody.data;
        const perms = await getPerms(book_id, {
            user_id: req.session!.user?.id,
            //ui_token_id: req.session.anonymous?.tokens[0]?.id, //only by user!
        });

        if (!perms.pACL) {
            res.status(400).json({
                error: "Not allowed to invite to this book",
            });
            return;
        }

        // Get potentialCollaborator and his status in the book
        let invitedUser;
        try {
            invitedUser = await service.user.getByEmail(email);
        } catch (error) {
            if (error instanceof Error && error.name === "NotFoundError") {
                res.status(400).json({
                    error: "User not found",
                });
                return;
            }
            res.status(500).json({
                error: "Failed to get user",
                details: (error as Error).message,
            });
            return;
        }

        // Check if already in book or already invited
        // invites are allowed every 2hours
        const alreadyCollaborator = await isAlreadyCollaborator(
            invitedUser.id,
            book_id,
        );

        if (alreadyCollaborator) {
            res.status(400).json({
                error: "Invited user is already a collaborator",
            });
            return;
        }

        // Check if any standing invites
        const pendingInvites =
            await service.collaboratorsInvite.getPendingByBookIdAndUserId(
                book_id,
                invitedUser.id,
            );

        if (pendingInvites.length > 0) {
            //Check all invites older than 2 hours
            const now = new Date();
            const twoHoursAgo = new Date(now.getTime() - 1000 * 60 * 60 * 2);

            const validInvites = pendingInvites.filter((invite) => {
                return invite.created_at > twoHoursAgo;
            });

            if (validInvites.length > 0) {
                res.status(400).json({
                    error: "Invite is still pending for this user you may try again later",
                    details:
                        "Invite is still pending for this user you may try again later",
                });
                return;
            }
        }

        // Add db entry
        const invite = await service.collaboratorsInvite.insert({
            book_id,
            user_id: invitedUser.id,
            invited_by: perms.entity_id,
        });

        // Get book for email information
        const [book, user] = await Promise.all([
            service.book.getById(book_id),
            service.user.getById(perms.entity_id),
        ]);

        const token = await signCollaborationRequestToken(invite);

        try {
            const emailToSend = new CollaboratorRequestEmail(
                invitedUser.email,
                book.title,
                book.id,
                user.email,
                token,
            );
            await sendEmail(emailToSend);
            res.json({
                success: true,
                invite,
            });
        } catch (error) {
            res.status(500).json({
                error: "Failed to send email",
                details: (error as Error).message,
            });
        }
    },
);

// check if row for user exist in permissions table
async function isAlreadyCollaborator(user_id: number, book_id: number) {
    try {
        await service.permission.getByEntityAndResource(
            user_id,
            book_id,
            ENTITY_USER,
            RESOURCE_BOOK,
        );
    } catch (error) {
        if (error instanceof NotFoundError) {
            return false;
        }
        throw error;
    }
    return true;
}
