import { unknownErrorToResponse } from "lib/errorparsing";
import { cookies } from "next/headers";
import { NextRequest, NextResponse } from "next/server";
import { UserInfoResponse } from "openid-client";

import { loadSSOProvider } from "../../page";
import { CookieSchema, getAccessToken, getUserInfo, Token } from "../oidc";

/** Callback or redirect url
 * which is used to handle the response from the
 * sso provider.
 *
 * Sometimes also called redirect uri.
 */
export async function GET(req: NextRequest) {
    // check if cookie is set
    const cookieStore = cookies();
    const cookie = cookieStore.get("sso_login");
    if (!cookie) {
        return NextResponse.json(
            {
                error: "Bad Request",
                details: "No cookie found for sso flow. Can't continue!",
            },
            {
                status: 400,
            },
        );
    }

    // Parse the cookie
    let data;
    try {
        data = JSON.parse(atob(cookie.value));
    } catch (e) {
        return unknownErrorToResponse("Failed to parse cookie", e, 400);
    }

    // Validate the cookie
    const pCookie = CookieSchema.safeParse(data);
    if (pCookie.success == false) {
        return NextResponse.json(
            { error: "Invalid cookie value", details: pCookie.error },
            {
                status: 400,
            },
        );
    }

    // Get provider by id
    const providers = await loadSSOProvider();
    const provider = providers.find(
        (p) => p.providerId === pCookie.data.providerId,
    );
    if (!provider) {
        return NextResponse.json(
            { error: "Invalid provider", details: "Provider not found" },
            {
                status: 400,
            },
        );
    }

    // Get the access token
    let token: Token;
    try {
        token = await getAccessToken(
            new URL(req.nextUrl.toString()),
            pCookie.data,
            provider,
        );
    } catch (e: unknown) {
        return unknownErrorToResponse("Failed to get access token", e, 400);
    }

    // Get user info
    let userInfo: UserInfoResponse;
    try {
        userInfo = await getUserInfo(token, provider);
    } catch (e: unknown) {
        return unknownErrorToResponse("Failed to get user info", e, 400);
    }

    return NextResponse.json({ userInfo });
    //TODO:
    // - Create user in database or check if user exists

    // Delete cookie, to cleanup after ourselves
    cookieStore.delete("sso_login");

    // Redirect to the home page
    // return NextResponse.redirect("/");
}
