import fs from "fs/promises";
import path from "path";
import { parse } from "yaml";
import z from "zod";

import { config } from "@snip/config/server";

import { BackToLoginBtn, SSO } from "./component";

import { Form, FormHeader } from "components/form";

/** YAML FILE schema */
const SSOProviderSchema = z.object({
    providers: z.preprocess(
        (val) => (val === null ? [] : val),
        z.array(
            z.object({
                // Name of the provider
                name: z.string(),

                // Description of the provider
                description: z.string(),

                // Image inside the config dir
                img: z.string(),

                // An identifier used to match the provider requests to the callback
                // should be url safe i.e. no special characters or spaces
                providerId: z.string().regex(/^[a-zA-Z0-9_-]+$/),

                // Only oidc /auth 2 is supported for now
                type: z.literal("oidc"),

                // Client ID and secret to authenticate client with the provider
                server: z.string().url(),
                clientId: z.string(),
                clientSecret: z.string(),
            }),
        ),
    ),
});

export type SSOProviders = z.infer<typeof SSOProviderSchema>["providers"];
export type SSOProvider = SSOProviders[0];

export default async function LoginSSO() {
    // Get available sso providers from yml file
    let providers = await loadSSOProvider();
    providers = await parseSSOProviderForUi(providers);

    if (providers.length === 0) {
        return (
            <div className="d-flex flex-column flex-grow-1 justify-content-center align-items-center w-100">
                <div>
                    <p>
                        This Deployment does not have any SSO providers
                        configured. Please contact your administrator.
                    </p>
                    <div className="d-flex align-items-center justify-content-center flex-column">
                        <BackToLoginBtn />
                    </div>
                </div>
            </div>
        );
    }

    return (
        <div className="d-flex flex-column flex-grow-1 justify-content-center align-items-center">
            <ProvidersForm providers={providers} />
        </div>
    );
}

async function ProvidersForm({ providers }: { providers: SSOProviders }) {
    return (
        <Form>
            <FormHeader>Please choose Single-Sign-On provider</FormHeader>

            <div className="d-flex align-items-center text-center mb-3 flex-column gap-2">
                {
                    // Render the SSO providers
                    providers.map((provider) => (
                        <SSO key={provider.providerId} {...provider} />
                    ))
                }
            </div>

            <div className="d-flex align-items-center justify-content-center flex-column">
                <BackToLoginBtn />
            </div>
        </Form>
    );
}

export async function loadSSOProvider(): Promise<SSOProviders> {
    const source = path.join(config.CONFIG_DIR, "sso_provider.yml");

    // check if file exists and create it if not
    if (!(await exists(source))) {
        console.warn(
            `File ${source} does not exist, generating empty providers file!`,
        );
        await fs.mkdir(path.dirname(source), {
            recursive: true,
        });
        await await fs.writeFile(source, EXAMPLE_FILE);
    }

    // Read the file
    const fileContent = await fs.readFile(source, "utf-8");

    // Parse yaml
    let yaml;
    try {
        yaml = parse(fileContent);
    } catch (e) {
        console.error(e);
        return [];
    }

    // Validate the yaml
    const data = SSOProviderSchema.safeParse(yaml);

    if (data.success == false) {
        console.error(data.error);
        return [];
    }

    return data.data.providers;
}

async function parseSSOProviderForUi(
    providers: SSOProviders,
): Promise<Omit<Omit<SSOProviders, "clientId">, "clientSecret">> {
    // Check if img paths are valid and load image as data uri
    for (const provider of providers) {
        const imgPath = path.join(config.CONFIG_DIR, provider.img);
        if (!(await exists(imgPath))) {
            console.warn(
                `Image file ${imgPath} does not exist! Using placeholder.`,
            );
            provider.img = "/img/favicon/favicon-128x128-light.png";
            continue;
        }

        const img = await fs.readFile(imgPath);
        provider.img = `data:image/png;base64,${img.toString("base64")}`;

        provider.clientId = "redacted";
        provider.clientSecret = "redacted";
    }
    return providers;
}

async function exists(f: string) {
    try {
        await fs.stat(f);
        return true;
    } catch {
        return false;
    }
}

const EXAMPLE_FILE = `providers:
# Example SSO providers configuration file
# - name: "Helmholz ID"
#   img: "helmholz.png" # -> /config/helmholz.png
#   description: "Login using the Helmholz AAI, allows a number of different institutions to authenticate."
#   
#   # Provider id, used to match the provider to the callback
#   # Client ID and secret to authenticate client with the provider
#   type: "oidc"
#   providerId: "helmholz"
#   server: "https://login.helmholtz.de/oauth2"
#   clientId: "123"
#   clientSecret: "123"
`;
