import Link from "next/link";

import { allKnownSchemas } from "./json/route";

import styles from "./schema.module.scss";

/** List all snip type schemas */
export default async function Page() {
    return (
        <div className="container-fluid flex-column justify-content-center align-items-center gap-3">
            <div className="container mt-3 col-xl-8 col-sm-12">
                <h1>Schemas</h1>
                <div>
                    Below, you&apos;ll find the schemas for all registered
                    snippets. These can be used to upload and visualize data on
                    our platform. Each snippet must include the required fields:{" "}
                    <code>type</code>,<code>book_id</code>, and{" "}
                    <code>data</code>. Optionally, a<code>view</code> field can
                    be provided to define the default visual representation,
                    though this can be adjusted when placing the snippet on a
                    page.
                </div>
                <div>
                    Snippet types are organized into namespaces, which are used
                    to group related snippets. For example, snippets related to
                    a specific measuring device or those associated with a
                    particular workgroup are categorized under relevant
                    namespaces.
                </div>
            </div>
            <div className="container mt-3 col-xl-8 col-sm-12">
                <h2>Available snippet types</h2>
                <div>
                    The table below lists all available snippet types,
                    automatically generated based on the registered snippets
                    within each namespace.
                </div>
                <SchemasTables />
            </div>
        </div>
    );
}

function SchemasTables() {
    const schemas = allKnownSchemas();
    schemas.delete("base");

    // Group by namespace i.e. everything before the last /
    const namespaces = new Map<string, [string, string | undefined][]>();
    for (const [key, _] of schemas.entries()) {
        const namespace = key.split("/").slice(0, -1).join("/");
        if (!namespaces.has(namespace)) {
            namespaces.set(namespace, []);
        }
        try {
            namespaces
                .get(namespace)
                ?.push([key, schemas.get(key)?.description]);
        } catch (e) {
            console.error(e);
        }
    }

    return (
        <div className={styles.schemas}>
            <figure>
                <figcaption>General snippets</figcaption>
                <table className={styles.table + " table"}>
                    <thead>
                        <tr>
                            <th>Type</th>
                            <th>Description</th>
                        </tr>
                    </thead>
                    <tbody>
                        <NamespaceRows
                            rows={namespaces.get("") || []}
                            namespace={undefined}
                        />
                    </tbody>
                </table>
            </figure>
            {namespaces.size > 1 && (
                <figure>
                    <figcaption>External snippets</figcaption>

                    <table className={styles.table}>
                        <thead>
                            <tr>
                                <th className={styles.type}>Type</th>
                                <th>Description</th>
                            </tr>
                        </thead>
                        <tbody>
                            {Array.from(namespaces.entries()).map(
                                ([namespace, rows]) => {
                                    if (namespace === "") {
                                        return null;
                                    }
                                    return (
                                        <NamespaceRows
                                            key={namespace}
                                            namespace={namespace}
                                            rows={rows}
                                        />
                                    );
                                },
                            )}
                        </tbody>
                    </table>
                </figure>
            )}
        </div>
    );
}

function NamespaceRows({
    namespace,
    rows,
}: {
    namespace: string | undefined;
    rows: [string, string | undefined][];
}) {
    if (rows.length === 0) {
        return null;
    }

    return (
        <>
            <tr>
                {namespace ? (
                    <th rowSpan={rows.length} scope="rowgroup">
                        {namespace}
                    </th>
                ) : null}
                <td scope="row">
                    <Link href={`./schemas/json/${rows[0]![0]}`}>
                        {rows[0]![0]}
                    </Link>
                </td>
                <td>{rows[0]![1] || "No description available"}</td>
            </tr>
            {rows.slice(1).map(([key, des]) => (
                <tr key={key}>
                    <td scope="row">
                        <Link href={`./schemas/json/${key}`}>{key}</Link>
                    </td>
                    <td>{des || "No description available"}</td>
                </tr>
            ))}
        </>
    );
}
