import { CollaboratorsInviteService } from "../collaborators";
import { BookData, UserData } from "../../types";
import { pool_permissions } from "../../sql.connection";
import { bookService } from "../book";
import { userService } from "../user";
import {
    InviteAlreadyAcceptedError,
    InviteExpiredError,
    NotFoundError,
} from "../../errors";

describe("CollaboratorInviteService", () => {
    let service: CollaboratorsInviteService;
    let consoleSpy;
    let bookData: BookData;
    let userData: UserData[];

    beforeAll(async () => {
        service = new CollaboratorsInviteService();

        await pool_permissions.query("DELETE FROM collaboration_invites");

        // Insert two users
        const u1 = await userService.insert({
            email: "owner@test.de",
            hashedPassword: "test",
        });
        const u2 = await userService.insert({
            email: "invite@test.de",
            hashedPassword: "test",
        });
        userData = [u1, u2];

        bookData = await bookService.insert({
            title: "test",
            owner_user_id: u1.id,
        });
    });

    afterAll(async () => {
        await bookService.delete(bookData.id);
        await userService.delete(userData[0]!.id);
        await userService.delete(userData[1]!.id);
    });

    describe("insert", () => {
        it("insert", async () => {
            const invite = await service.insert({
                book_id: bookData.id,
                user_id: userData[1]!.id,
                invited_by: userData[0]!.id,
            });

            expect(invite).toMatchObject({
                book_id: bookData.id,
                user_id: userData[1]!.id,
                invited_by: userData[0]!.id,
                user_email: userData[1]!.email,
                invited_by_email: userData[0]!.email,
            });

            // Check expire at is set and approx 48 hours in the future
            expect(invite.expire_at).toBeInstanceOf(Date);
            expect(invite.expire_at.getTime()).toBeGreaterThan(Date.now());
            expect(invite.expire_at.getTime()).toBeLessThan(
                Date.now() + 1000 * 60 * 60 * 48,
            );
            expect(invite.expire_at.getTime()).toBeGreaterThan(
                Date.now() + 1000 * 60 * 60 * 47,
            );

            await service.delete(invite.id);
        });
    });

    describe("get", () => {
        let pending: any;
        let accepted: any;
        beforeAll(async () => {
            pending = await service.insert({
                book_id: bookData.id,
                user_id: userData[1]!.id,
                invited_by: userData[0]!.id,
            });

            accepted = await service.insert({
                book_id: bookData.id,
                user_id: userData[1]!.id,
                invited_by: userData[0]!.id,
            });
            await pool_permissions.query(
                `UPDATE collaboration_invites SET accepted_at = NOW() WHERE id = ?`,
                [accepted.id],
            );
        });
        afterAll(async () => {
            await service.delete(pending.id);
            await service.delete(accepted.id);
        });

        it("getById", async () => {
            const result = await service.getById(pending.id);

            expect(result).toMatchObject(pending);

            expect(result.accepted_at).toBeNull();

            // throw on not found
            await expect(service.getById(-1)).rejects.toThrow(NotFoundError);
        });

        it("getByUserId", async () => {
            const result = await service.getByUserId(userData[1]!.id);

            expect(result[0]).toMatchObject({
                ...pending,
            });
            expect(result.length).toBe(2);
        });

        it("getByBookId", async () => {
            const result = await service.getByBookId(bookData.id);

            expect(result[0]).toMatchObject({
                ...pending,
            });
            expect(result.length).toBe(2);
        });

        it("getPendingByBookId", async () => {
            const result = await service.getPendingByBookId(bookData.id);

            expect(result[0]).toMatchObject(pending);
            expect(result[0]!.accepted_at).toBeNull();
            expect(result.length).toBe(1);
        });

        it("getPendingByUserId", async () => {
            const result = await service.getPendingByUserId(userData[1]!.id);

            expect(result[0]).toMatchObject(pending);
            expect(result[0]!.accepted_at).toBeNull();
            expect(result.length).toBe(1);
        });

        it("getPendingByBookIdAndUserId", async () => {
            const result = await service.getPendingByBookIdAndUserId(
                bookData.id,
                userData[1]!.id,
            );

            expect(result[0]).toMatchObject(pending);
            expect(result[0]!.accepted_at).toBeNull();
            expect(result.length).toBe(1);
        });
    });

    describe("accept", () => {
        let pending: any;
        beforeAll(async () => {
            pending = await service.insert({
                book_id: bookData.id,
                user_id: userData[1]!.id,
                invited_by: userData[0]!.id,
            });
        });

        beforeEach(async () => {
            /**
             */
            await pool_permissions.query(
                `UPDATE collaboration_invites SET accepted_at = NULL WHERE id = ?`,
                [pending.id],
            );
            // Delete acl entry
            await pool_permissions.query(
                `DELETE FROM acl WHERE resource_id = ? AND entity_id = ?`,
                [bookData.id, userData[1]!.id],
            );
        });

        afterAll(async () => {
            await service.delete(pending.id);
        });

        it("acceptById", async () => {
            const result = await service.acceptById(pending.id);

            expect(result).toMatchObject({
                ...pending,
                accepted_at: expect.any(Date),
            });
            expect(result.accepted_at!.getTime()).toBeGreaterThan(
                Date.now() - 1000,
            );
        });

        it("acceptById should throw if invite does not exist", async () => {
            await expect(service.acceptById(-1)).rejects.toThrow(NotFoundError);
        });

        it("acceptById should throw if invite is already accepted", async () => {
            await pool_permissions.query(
                `UPDATE collaboration_invites SET accepted_at = NOW() WHERE id = ?`,
                [pending.id],
            );

            await expect(service.acceptById(pending.id)).rejects.toThrow(
                InviteAlreadyAcceptedError,
            );
        });

        it("acceptById should throw if invite is expired", async () => {
            await pool_permissions.query(
                `UPDATE collaboration_invites SET expire_at = NOW() WHERE id = ?`,
                [pending.id],
            );

            await expect(service.acceptById(pending.id)).rejects.toThrow(
                InviteExpiredError,
            );
        });
    });
});
