import { NotFoundError } from "../../errors";
import { pool_permissions } from "../../sql.connection";
import { GroupData, PartialBy } from "../../types";
import { GroupService } from "../group";

describe("GroupService", () => {
    let service: GroupService;
    let consoleSpy;
    let defaultGroupData: PartialBy<GroupData, "id">;
    let user_ids: number[] = [];

    async function deleteCleanup() {
        await pool_permissions.query(
            "DELETE FROM users_groups WHERE user_id in (select user_id from users where email like '%@group.de')",
        );
        await pool_permissions.query(
            "DELETE FROM groups where name like 'test.group'",
        );
        await pool_permissions.query(
            "DELETE FROM users WHERE email like '%@group.de'",
        );
    }

    beforeAll(async () => {
        service = new GroupService();

        defaultGroupData = {
            name: "test.group",
            description: "Test description",
        };

        //Insert a user
        await deleteCleanup();
        await pool_permissions
            .query(
                "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES ('test123@group.de', 1, 'test')",
            )
            .then((res) => {
                user_ids.push(res.insertId);
            });
    });

    afterAll(async () => {
        await deleteCleanup();
    });

    describe("insert", () => {
        it("should insert a new group", async () => {
            const groupData = await service.insert(
                structuredClone(defaultGroupData),
            );
            expect_same(groupData, defaultGroupData);
            expect(groupData.id).toBeDefined();
            expect(groupData.created).toBeDefined();
            expect(groupData.last_updated).toBeDefined();
        });
    });

    describe("getById|getByIds", () => {
        it("should retrieve a group by its ID(s)", async () => {
            const groupData = await service.insert(
                structuredClone(defaultGroupData),
            );
            const groupData2 = await service.insert(
                structuredClone(defaultGroupData),
            );
            const retrievedGroupData = await service.getById(groupData.id);
            expect_same(retrievedGroupData, groupData, true);

            const retrievedGroupData2 = await service.getByIds([
                groupData.id,
                groupData2.id,
            ]);
            expect(retrievedGroupData2.length).toBe(2);
            expect_same(retrievedGroupData2[0]!, groupData, true);
            expect_same(retrievedGroupData2[1]!, groupData2, true);
        });

        it("should throw an error if the group(s) not found", async () => {
            await expect(service.getById(-1)).rejects.toThrow();

            await expect(service.getByIds([-1])).rejects.toThrow();
        });
    });

    describe("update", () => {
        it("should update a group", async () => {
            const groupData = await service.insert(
                structuredClone(defaultGroupData),
            );
            const updatedGroupData = await service.update({
                ...groupData,
                description: "Updated description",
            });
            expect_same(updatedGroupData, {
                ...groupData,
                description: "Updated description",
            });
            expect(updatedGroupData.id).toBe(groupData.id);
            expect(updatedGroupData.created).toBe(groupData.created);
            expect(updatedGroupData.last_updated!.getTime()).toBeGreaterThan(
                groupData.last_updated!.getTime(),
            );
        });

        it("should throw an error if the group not found", async () => {
            await expect(
                service.update({
                    ...defaultGroupData,
                    id: -1,
                }),
            ).rejects.toThrow();
        });
    });

    describe("delete", () => {
        it("should delete a group", async () => {
            const groupData = await service.insert(
                structuredClone(defaultGroupData),
            );
            await service.delete(groupData.id);
            await expect(service.getById(groupData.id)).rejects.toThrow();
        });

        it("should throw an error if the group not found", async () => {
            await expect(service.delete(-1)).rejects.toThrow();
        });
    });

    describe("getByUserID|getGroupIDsByUserID", () => {
        it("should retrieve groups by user ID", async () => {
            const groupData = await service.insert(
                structuredClone(defaultGroupData),
            );
            const groupData2 = await service.insert(
                structuredClone(defaultGroupData),
            );
            const groupData3 = await service.insert(
                structuredClone(defaultGroupData),
            );

            const user_id = await pool_permissions
                .query(
                    "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES (?, 1, 'test')",
                    ["test123111@group.de"],
                )
                .then((res) => res.insertId);

            await service.addMember(user_id, groupData.id);
            await service.addMember(user_id, groupData2.id);

            const retrievedGroupData = await service.getByUserId(user_id);
            expect(retrievedGroupData.length).toBe(2);
            expect_same(retrievedGroupData[0]!, groupData);
            expect_same(retrievedGroupData[1]!, groupData2);
        });

        it("should retrieve group IDs by user ID", async () => {
            const groupData = await service.insert(
                structuredClone(defaultGroupData),
            );
            const groupData2 = await service.insert(
                structuredClone(defaultGroupData),
            );
            const groupData3 = await service.insert(
                structuredClone(defaultGroupData),
            );
            const user_id = await pool_permissions
                .query(
                    "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES (?, 1, 'test')",
                    ["test1231121@group.de"],
                )
                .then((res) => res.insertId);

            await service.addMember(user_id, groupData.id);
            await service.addMember(user_id, groupData2.id);

            const retrievedGroupIDs =
                await service.getGroupIdsByUserId(user_id);
            expect(retrievedGroupIDs.length).toBeGreaterThanOrEqual(2);
            expect(retrievedGroupIDs).toContain(groupData.id);
            expect(retrievedGroupIDs).toContain(groupData2.id);
            expect(retrievedGroupIDs).not.toContain(groupData3.id);
        });
    });

    describe("search", () => {
        test.todo("add test");
    });

    /* -------------------------------------------------------------------------- */
    /*                                group members                               */
    /* -------------------------------------------------------------------------- */

    describe("members", () => {
        describe("addMember", () => {
            it("should add members", async () => {
                const groupData = await service.insert(
                    structuredClone(defaultGroupData),
                );

                for (const role of [
                    "invited",
                    "member",
                    "owner",
                    "moderator",
                ]) {
                    const user_id = await pool_permissions
                        .query(
                            "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES (?, 1, 'test')",
                            [`${role}@group.de`],
                        )
                        .then((res) => res.insertId);

                    const member = await service.addMember(
                        user_id,
                        groupData.id,
                        role as any,
                    );
                    expect(member.group_id).toBe(groupData.id);
                    expect(member.user_id).toBe(user_id);
                    //Date
                    expect(member.joined_at).toBeDefined();
                    expect(member.joined_at).toBeInstanceOf(Date);
                    //Role
                    expect(member.role).toBe(role);
                }
            });

            it("should overwrite role if already exists", async () => {
                const groupData = await service.insert(
                    structuredClone(defaultGroupData),
                );

                //Should already exist see above
                const user_id = await pool_permissions
                    .query(
                        "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES ('overwriteTest@group.de', 1, 'test')",
                    )
                    .then((res) => res.insertId);

                // Inital add
                const member = await service.addMember(user_id, groupData.id);
                expect(member.group_id).toBe(groupData.id);
                expect(member.user_id).toBe(user_id);
                expect(member.joined_at).toBeDefined();
                expect(member.joined_at).toBeInstanceOf(Date);
                expect(member.role).toBe("invited");

                await new Promise((resolve) => setTimeout(resolve, 2000));

                // Overwrite
                const member2 = await service.addMember(
                    user_id,
                    groupData.id,
                    "member",
                );

                expect(member2.group_id).toBe(groupData.id);
                expect(member2.user_id).toBe(user_id);
                expect(member2.joined_at).toBeDefined();
                expect(member2.joined_at).toBeInstanceOf(Date);
                expect(member2.role).toBe("member");
                //Date should update
                expect(member2.joined_at.getTime()).toBeGreaterThan(
                    member.joined_at.getTime(),
                );

                await new Promise((resolve) => setTimeout(resolve, 2000));

                // Overwrite
                const member3 = await service.addMember(
                    user_id,
                    groupData.id,
                    "owner",
                );
                expect(member3.group_id).toBe(groupData.id);
                expect(member3.user_id).toBe(user_id);
                expect(member3.joined_at).toBeDefined();
                expect(member3.joined_at).toBeInstanceOf(Date);
                expect(member3.role).toBe("owner");
                //Date should not update!
                expect(member3.joined_at.getTime()).toBe(
                    member2.joined_at.getTime(),
                );

                await new Promise((resolve) => setTimeout(resolve, 2000));

                // Overwrite
                const member4 = await service.addMember(
                    user_id,
                    groupData.id,
                    "moderator",
                );
                expect(member4.group_id).toBe(groupData.id);
                expect(member4.user_id).toBe(user_id);
                expect(member4.joined_at).toBeDefined();
                expect(member4.joined_at).toBeInstanceOf(Date);
                expect(member4.role).toBe("moderator");
                //Date should not update!
                expect(member4.joined_at.getTime()).toBe(
                    member3.joined_at.getTime(),
                );
            }, 10000);
        });

        describe("getMember(s)", () => {
            it("should retrieve a member by user ID and group ID", async () => {
                const groupData = await service.insert(
                    structuredClone(defaultGroupData),
                );

                const user_id = await pool_permissions
                    .query(
                        "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES ('testgetMember@group.de', 1, 'test')",
                    )
                    .then((res) => res.insertId);

                const members = await service.getMembers(groupData.id);
                expect(members.length).toBe(0);

                const member = await service.addMember(user_id, groupData.id);
                expect(member.group_id).toBe(groupData.id);
                expect(member.user_id).toBe(user_id);
                expect(member.joined_at).toBeDefined();
                expect(member.joined_at).toBeInstanceOf(Date);
                expect(member.role).toBe("invited");
                expect(member.email).toBe("testgetMember@group.de");

                const members2 = await service.getMembers(groupData.id);
                expect(members2.length).toBe(1);
                expect(members2[0]!.group_id).toBe(groupData.id);
                expect(members2[0]!.user_id).toBe(user_id);
                expect(members2[0]!.joined_at).toBeDefined();
                expect(members2[0]!.joined_at).toBeInstanceOf(Date);
                expect(members2[0]!.role).toBe("invited");

                //Overlap
                expect(members2[0]).toStrictEqual(member);
            });

            it("should throw not found error if member not found", async () => {
                await expect(service.getMember(-1, -1)).rejects.toThrow(
                    NotFoundError,
                );

                // Empty on getMembers
                const members = await service.getMembers(-1);
                expect(members.length).toBe(0);
            });
        });

        describe("isAllowedToEdit", () => {
            it("existing members", async () => {
                const groupData = await service.insert(
                    structuredClone(defaultGroupData),
                );

                const roles = {
                    name: ["invited", "member", "moderator", "owner"],
                    allowed: [false, false, true, true],
                };

                for (let i = 0; i < roles.name.length; i++) {
                    const role = roles.name[i];

                    const user_id = await pool_permissions
                        .query(
                            "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES (?, 1, 'test')",
                            [`${role}allowed@group.de`],
                        )
                        .then((res) => res.insertId);

                    const member = await service.addMember(
                        user_id,
                        groupData.id,
                        role as any,
                    );
                    expect(member.group_id).toBe(groupData.id);
                    expect(member.user_id).toBe(user_id);
                    //Date
                    expect(member.joined_at).toBeDefined();
                    expect(member.joined_at).toBeInstanceOf(Date);
                    //Role
                    expect(member.role).toBe(role);

                    // Check if allowed to edit
                    const allowed = await service.isAllowedToEdit(
                        user_id,
                        groupData.id,
                    );
                    expect(allowed).toBe(roles.allowed[i]);
                }
            });
            it("not existing members", async () => {
                const groupData = await service.insert(
                    structuredClone(defaultGroupData),
                );

                const allowed = await service.isAllowedToEdit(-1, groupData.id);
                expect(allowed).toBe(false);
            });
        });

        describe("removeMember", () => {
            it("should remove a member", async () => {
                const groupData = await service.insert(
                    structuredClone(defaultGroupData),
                );

                const user_id = await pool_permissions
                    .query(
                        "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES ('removemember@group.de', 1, 'test')",
                    )
                    .then((res) => res.insertId);

                const member = await service.addMember(user_id, groupData.id);
                expect(member.group_id).toBe(groupData.id);
                expect(member.user_id).toBe(user_id);
                expect(member.joined_at).toBeDefined();
                expect(member.joined_at).toBeInstanceOf(Date);

                await service.removeMember(user_id, groupData.id);
                await expect(
                    service.getMember(user_id, groupData.id),
                ).rejects.toThrow(NotFoundError);
            });

            it("should throw an error if the member not found", async () => {
                await expect(service.removeMember(-1, -1)).rejects.toThrow(
                    NotFoundError,
                );
            });
        });

        it("getRole", async () => {
            const groupData = await service.insert(
                structuredClone(defaultGroupData),
            );

            const user_id = await pool_permissions
                .query(
                    "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES ('getRole@group.de', 1, 'test')",
                )
                .then((res) => res.insertId);

            const member = await service.addMember(user_id, groupData.id);

            const getRole = await service.getRole(user_id, groupData.id);
            expect(getRole).toBe("invited");

            const getRole2 = await service.getRole(-1, -1);
            expect(getRole2).toBe(undefined);

            const getRole3 = await service.getRole(user_id, -1);
            expect(getRole3).toBe(undefined);

            const getRole4 = await service.getRole(-1, groupData.id);
            expect(getRole4).toBe(undefined);
        });
    });
});

function expect_same(
    A: PartialBy<GroupData, "id">,
    B: PartialBy<GroupData, "id">,
    times = false,
) {
    expect(A.name).toBe(B.name);
    expect(A.description).toBe(B.description);
    if (times) {
        expect(A.created!.getTime()).toBeCloseTo(B.created!.getTime(), -4);
        expect(A.last_updated!.getTime()).toBeCloseTo(
            B.last_updated!.getTime(),
            -4,
        );
    }
}
