import { UserService } from "../user";
import { PartialBy, UserData } from "../../types";
import { pool_permissions } from "../../sql.connection";

describe("UserService", () => {
    let service: UserService;
    let consoleSpy;
    let defaultUserData: PartialBy<UserData, "id">;

    beforeAll(async () => {
        service = new UserService();

        defaultUserData = {
            email: "test@users.de",
        };
    });

    afterAll(async () => {
        await pool_permissions.query(
            "DELETE FROM users_groups where user_id in (select user_id from users where email like '%@users.de')",
        );
        await pool_permissions.query(
            "DELETE FROM users where email like '%@users.de'",
        );
    });

    describe("insert", () => {
        it("should insert a new user", async () => {
            const userData = await service.insert(
                structuredClone({ ...defaultUserData, hashedPassword: "test" }),
            );
            expect_same(userData, defaultUserData);
            await pool_permissions.query(
                "DELETE FROM users where email like '%@users.de'",
            );
        });
    });

    describe("getById|getByIds", () => {
        it("should retrieve a single user by its ID(s)", async () => {
            const userData = await service.insert(
                structuredClone({ ...defaultUserData, hashedPassword: "test" }),
            );
            const retrievedUserData = await service.getById(userData.id);
            expect_same(retrievedUserData, userData, true, true);

            const userData2 = await service.insert(
                structuredClone({
                    ...defaultUserData,
                    hashedPassword: "test",
                    email: "test2@users.de",
                }),
            );
            const retrievedUserData2 = await service.getByIds([
                userData.id,
                userData2.id,
            ]);
            expect(retrievedUserData2.length).toBe(2);
            expect_same(retrievedUserData2[0]!, userData, true, true);
            expect_same(retrievedUserData2[1]!, userData2, true, true);

            await pool_permissions.query(
                "DELETE FROM users where email like '%@users.de'",
            );
        });

        it("should throw an error if the user(s) not found", async () => {
            await expect(service.getById(-1)).rejects.toThrow();

            await expect(service.getByIds([-1])).rejects.toThrow();
        });
    });

    describe("update", () => {
        it("should update a user", async () => {
            const userData = await service.insert(
                structuredClone({ ...defaultUserData, hashedPassword: "test" }),
            );
            const updatedUserData = await service.update({
                ...userData,
                email: "test2@users.de",
            });

            expect_same(
                updatedUserData,
                {
                    ...userData,
                    email: "test2@users.de",
                },
                true,
                true,
            );
        });

        it("should throw an error if the user does not exist", async () => {
            await expect(
                service.update({
                    id: -1,
                    email: "teadsa@test.de",
                }),
            ).rejects.toThrow();
        });
    });

    describe("delete", () => {
        it("should delete a user", async () => {
            const userData = await service.insert(
                structuredClone({ ...defaultUserData, hashedPassword: "test" }),
            );
            await service.delete(userData.id);
            await expect(service.getById(userData.id)).rejects.toThrow();
        });

        it("should throw an error if the user does not exist", async () => {
            await expect(service.delete(-1)).rejects.toThrow();
        });
    });

    describe("signup|login|verify|updatePassword", () => {
        it("signup a new user with a password", async () => {
            const userData = await service.signup("test44@users.de", "test");
            expect(userData.email).toBe("test44@users.de");
            expect(userData.id).toBeDefined();
        });
        it("signup should reject if the email is already taken", async () => {
            await expect(
                service.signup("test44@users.de", "test"),
            ).rejects.toThrow();
        });
        it("verify a users email", async () => {
            const succ = await service.verify("test44@users.de");
            expect(succ).toBe(true);
        });
        it("verify should throw if the email is not found", async () => {
            await expect(service.verify("test76@users.de")).rejects.toThrow();
        });
        it("login a user", async () => {
            const succ = await service.login("test44@users.de", "test");
            expect(succ).toBe(true);
        });
        it("login should be false if the email is not found or pw is wrong", async () => {
            const succ = await service.login("test44@users.de", "test12");
            expect(succ).toBe(false);

            const succ2 = await service.login("test76@users.de", "test");
            expect(succ2).toBe(false);
        });
        it("updatePassword a users password", async () => {
            const succ = await service.updatePassword(
                "test44@users.de",
                "test22",
            );
            expect(succ).toBe(true);

            const succ2 = await service.login("test44@users.de", "test22");
            expect(succ2).toBe(true);
        });
        it("updatePassword should throw if the email is not found", async () => {
            await expect(
                service.updatePassword("test76@users.de", "test"),
            ).rejects.toThrow();
        });
    });

    describe("getByEmail", () => {
        it("should retrieve a user by its email", async () => {
            const userData = await service.insert(
                structuredClone({ ...defaultUserData, hashedPassword: "test" }),
            );
            const retrievedUserData = await service.getByEmail(userData.email);
            expect_same(retrievedUserData, userData, true, true);
        });

        it("should throw an error if the user not found", async () => {
            await expect(
                service.getByEmail("test76@users.de"),
            ).rejects.toThrow();
        });
    });

    describe("isAdmin", () => {
        it("should return true if the user is an admin", async () => {
            const userData = await service.insert(
                structuredClone({
                    ...defaultUserData,
                    hashedPassword: "test",
                    email: "admin@users.de",
                }),
            );
            await pool_permissions.query(
                `INSERT INTO users_groups (user_id, group_id) VALUES (${userData.id}, 1)`,
            );
            const isAdmin = await service.isAdmin(userData.id);
            expect(isAdmin).toBe(true);
        });

        it("should return false if the user is not an admin", async () => {
            const userData = await service.insert(
                structuredClone({
                    ...defaultUserData,
                    hashedPassword: "test",
                    email: "noadmin@users.de",
                }),
            );
            const isAdmin = await service.isAdmin(userData.id);
            expect(isAdmin).toBe(false);
        });
    });

    describe("search", () => {
        it("should return an array of users that match the search query", async () => {
            const userData1 = await service.search("users");
            expect(userData1.length).toBeGreaterThanOrEqual(1);
        });
    });
    describe("getPasswordHash", () => {
        it("should return the hashed password of a user", async () => {
            const hash = await service.getPasswordHash("test@users.de");
            expect(hash).toBe("test");
        });
    });
});

function expect_same(
    received: PartialBy<UserData, "id">,
    expected: PartialBy<UserData, "id">,
    id = false,
    time = false,
) {
    expect(received.email).toBe(expected.email);
    expect(received.id).toBeDefined();
    expect(received.created).toBeDefined();
    expect(received.last_updated).toBeDefined();

    if (id) {
        expect(received.id).toBe(expected.id);
    }
    if (time) {
        expect(received.created!.getTime()).toBeCloseTo(
            expected.created!.getTime(),
            -4,
        );
        expect(received.last_updated!.getTime()).toBeCloseTo(
            expected.last_updated!.getTime(),
            -4,
        );
    }
}
