import { UserConfigService } from "../userConfig";
import { PartialBy, UserConfigData } from "../../types";
import { pool_permissions } from "../../sql.connection";

describe("UserConfigService", () => {
    let service: UserConfigService;
    let defaultUserConfigData: PartialBy<UserConfigData, "id">;
    let consoleSpy;

    async function deleteCleanup() {
        await pool_permissions.query(
            "DELETE FROM user_config WHERE user_id = (SELECT id FROM users WHERE email = 'test@userconfig.de')",
        );
        await pool_permissions.query(
            "DELETE FROM users WHERE email = 'test@userconfig.de'",
        );
    }

    beforeAll(async () => {
        service = new UserConfigService();

        // Insert a user
        await deleteCleanup();
        const u_id = await pool_permissions
            .query(
                "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES (?, 1, 'test')",
                ["test@userconfig.de"],
            )
            .then((res) => {
                return res.insertId;
            });

        defaultUserConfigData = {
            user_id: u_id,
            text_font: "Arial",
        };
    });
    afterAll(async () => {
        await deleteCleanup();
    });

    describe("getByIds", () => {
        it("not implemented", async () => {
            await expect(service.getByIds([-1])).rejects.toThrow(
                "Method not implemented.",
            );
        });
    });

    describe("insert|delete", () => {
        it("should insert a new user configuration", async () => {
            const userConfigData = await service.insert(
                structuredClone(defaultUserConfigData),
            );
            expect(userConfigData.id).toBeDefined();

            //cleanup
            await expect(service.delete(userConfigData.id)).resolves.toBe(true);
        });
        it("should throw an error if the user does not exist", async () => {
            await expect(
                service.insert({ ...defaultUserConfigData, user_id: -1 }),
            ).rejects.toThrow();

            await expect(service.delete(-1)).rejects.toThrow();
        });
        it("should throw an error if the user configuration already exists", async () => {
            const userConfigData = await service.insert(
                structuredClone(defaultUserConfigData),
            );
            await expect(
                service.insert(structuredClone(defaultUserConfigData)),
            ).rejects.toThrow();

            //cleanup
            await expect(service.delete(userConfigData.id)).resolves.toBe(true);
        });
        it("should throw an error if the user does not exist", async () => {
            await expect(
                service.insert(
                    structuredClone({
                        ...defaultUserConfigData,
                        user_id: 999,
                    }),
                ),
            ).rejects.toThrow();
        });
    });

    describe("update|updateByUserId", () => {
        it("should partially update a user configuration", async () => {
            const config = await service.insert(
                structuredClone(defaultUserConfigData),
            );

            //normal update by id
            const userConfigData = await service.update({
                text_font: "Caveat",
                id: config.id,
                pen_size: 20,
            });
            expect(userConfigData.text_font).toBe("Caveat");
            expect(userConfigData.pen_size).toBe(20);

            //update by user_id
            const userConfigData2 = await service.updateByUserId({
                text_font: "Arial",
                user_id: config.user_id,
                pen_size: 40,
            });

            expect(userConfigData2.text_font).toBe("Arial");
            expect(userConfigData2.pen_size).toBe(40);

            //cleanup
            await expect(service.delete(userConfigData.id)).resolves.toBe(true);
        });

        it("should throw an error if the user configuration does not exist", async () => {
            await expect(
                service.update({ ...defaultUserConfigData, id: -1 }),
            ).rejects.toThrow();

            await expect(
                service.updateByUserId({
                    ...defaultUserConfigData,
                    user_id: -1,
                }),
            ).rejects.toThrow();
        });

        it("should throw an error if no update fields are provided", async () => {
            const userConfigData = await service.insert(
                structuredClone(defaultUserConfigData),
            );
            await expect(
                service.update({ id: userConfigData.id }),
            ).rejects.toThrow();

            await expect(
                service.updateByUserId({ user_id: userConfigData.user_id }),
            ).rejects.toThrow();

            //cleanup
            await expect(service.delete(userConfigData.id)).resolves.toBe(true);
        });

        it("should throw if the font does not exist", async () => {
            const userConfigData = await service.insert(
                structuredClone(defaultUserConfigData),
            );
            await expect(
                service.update({ ...userConfigData, text_font: "not_a_font" }),
            ).rejects.toThrow();

            await expect(
                service.updateByUserId({
                    ...userConfigData,
                    text_font: "not_a_font",
                }),
            ).rejects.toThrow();

            //cleanup
            await expect(service.delete(userConfigData.id)).resolves.toBe(true);
        });
    });

    describe("getById|getByUserId", () => {
        it("should retrieve a user configuration by its ID(s)", async () => {
            const userConfigData = await service.insert(
                structuredClone(defaultUserConfigData),
            );
            const retrievedUserConfigData = await service.getById(
                userConfigData.id,
            );
            expect(retrievedUserConfigData.id).toBe(userConfigData.id);

            const retrievedUserConfigData2 = await service.getByUserId(
                userConfigData.user_id,
            );
            expect(retrievedUserConfigData2.id).toBe(userConfigData.id);

            //cleanup
            await expect(service.delete(userConfigData.id)).resolves.toBe(true);
        });

        it("should throw an error if the user configuration does not exist", async () => {
            await expect(service.getById(-1)).rejects.toThrow();

            await expect(service.getByUserId(-1)).rejects.toThrow();
        });
    });

    describe("getAll", () => {
        it("should retrieve all user configurations", async () => {
            const userConfigData = await service.insert(
                structuredClone(defaultUserConfigData),
            );
            const retrievedUserConfigData = await service.getAll();
            expect(retrievedUserConfigData.length).toBeGreaterThanOrEqual(1);

            //cleanup
            await expect(service.delete(userConfigData.id)).resolves.toBe(true);
        });
    });

    describe("search", () => {
        it("should return an array of configs", async () => {
            await expect(service.search("asd")).rejects.toThrow();
        });
    });
});
