import {
    BackgroundTypeData,
    BaseData,
    BaseView,
    BlobData,
    BookData,
    BookDataResolved,
    BookOwner,
    Entity,
    FileData,
    FolderData,
    GroupData,
    GroupMemberData,
    ID,
    PageData,
    PartialBy,
    PermissionACLData,
    PermissionACLDataResolved,
    Resource,
    SnipData,
    UserConfigData,
    UserData,
} from "./types";

/** We more or less use the strategy pattern to
 * create a database strategy which allows to interact with
 * different tables.
 *
 *
 * see https://en.wikipedia.org/wiki/Strategy_pattern
 */

export interface TableStrategy<T> {
    getAll: () => Promise<(T & { id: ID })[]>;
    getById: (id: ID) => Promise<T & { id: ID }>;
    getByIds: (ids: ID[]) => Promise<(T & { id: ID })[]>;
    insert: (data: T) => Promise<T & { id: ID }>;
    update: (data: T & { id: ID }) => Promise<T & { id: ID }>;
    delete: (id: ID) => Promise<boolean>;
    upsert?: (data: T) => Promise<T & { id: ID }>;
    search: (query: string) => Promise<T[]>;
}

export interface UserStrategy extends TableStrategy<UserData> {
    insert: (data: UserData & { hashedPassword: string }) => Promise<UserData>;
    login: (email: string, password: string) => Promise<boolean>;
    loginById: (id: ID, password: string) => Promise<boolean>;
    verify: (email: string) => Promise<boolean>;
    getByEmail: (email: string) => Promise<UserData>;
    signup: (email: string, password: string) => Promise<UserData>;
    isAdmin: (user_id: ID) => Promise<boolean>;
    search: (search: string) => Promise<UserData[]>;
    getPasswordHash: (email: string) => Promise<string>;
    updatePassword: (email: string, password: string) => Promise<boolean>;
    updateLoginInfo: (user_id: ID, ip: string) => Promise<boolean>;
}

export interface GroupStrategy extends TableStrategy<GroupData> {
    getByUserId: (user_id: ID) => Promise<GroupData[]>;
    getGroupIdsByUserId: (user_id: ID) => Promise<ID[]>;
    getMembers: (group_id: ID) => Promise<Required<GroupMemberData>[]>;
    getMember: (
        user_id: ID,
        group_id: ID,
    ) => Promise<Required<GroupMemberData>>;
    addMember: (
        user_id: ID,
        group_id: ID,
        role: GroupMemberData["role"],
    ) => Promise<Required<GroupMemberData>>;
    updateMember: (
        user_id: ID,
        group_id: ID,
        role: GroupMemberData["role"],
    ) => Promise<Required<GroupMemberData>>;
    removeMember: (user_id: ID, group_id: ID) => Promise<boolean>;
    isAllowedToEdit: (user_id: ID, group_id: ID) => Promise<boolean>;
    getRole: (
        user_id: ID,
        group_id: ID,
    ) => Promise<GroupMemberData["role"] | undefined>;
}

export interface PageStrategy extends TableStrategy<PageData> {
    getByBookId: (book_id: ID) => Promise<PageData[]>;
    getFirstByBookId: (book_id: ID) => Promise<PageData>;
    getLastByBookId: (book_id: ID) => Promise<PageData>;
    getCoverByBookId: (book_id: ID) => Promise<PageData>;
}

export interface BlobStrategy extends TableStrategy<BlobData> {}

export interface BookStrategy extends TableStrategy<BookData> {
    setOwner: (
        entity_id: ID,
        entity_type: "user" | "group",
        book_id: ID,
    ) => Promise<boolean>;
    setOwnerByName: (entity_name: string, book_id: ID) => Promise<BookOwner>;

    // Resolved data is data join with other tables to get the owner and the default background type
    getByIdResolved: (ids: ID) => Promise<BookDataResolved>;
    getByIdsResolved: (ids: ID[]) => Promise<BookDataResolved[]>;

    getByUserId(user_id: ID, pWrite?: boolean): Promise<BookData[]>;
}

export interface SnipStrategy
    extends TableStrategy<SnipData<BaseData, BaseView>> {
    getByPageId: (page_id: number) => Promise<SnipData<BaseData, BaseView>[]>;
    getQueuedByBookId: (
        book_id: number,
    ) => Promise<SnipData<BaseData, BaseView>[]>;
    placeOnPage: (
        snip_id: number,
        page_id: number,
    ) => Promise<SnipData<BaseData, BaseView>>;
    upsert: (
        snip_data: SnipData<BaseData, BaseView>,
    ) => Promise<SnipData<BaseData, BaseView> & { id: number }>;
    deleteNested: (snip_id: ID) => Promise<boolean>;
}

export interface UserConfigStrategy extends TableStrategy<UserConfigData> {
    getByUserId: (user_id: ID) => Promise<UserConfigData>;
    update: (
        data: Partial<UserConfigData> & { id: ID },
    ) => Promise<UserConfigData & { id: ID }>;
    updateByUserId: (
        data: Partial<UserConfigData> & { user_id: ID },
    ) => Promise<UserConfigData>;
}

export interface FolderStrategy extends TableStrategy<FolderData> {
    delete: (id: ID, user_id?: ID) => Promise<boolean>;
    move: (
        type: "folder" | "book",
        item: ID,
        moveTo: ID | "root",
        user_id: ID,
    ) => Promise<boolean>;
    getFilesByUserId: (user_id: ID) => Promise<FileData[]>;
}

export interface BackgroundTypeStrategy
    extends TableStrategy<BackgroundTypeData> {
    setBookDefault: (book_id: ID, background_type_id: ID) => Promise<true>;

    setPageById: (
        book_id: ID,
        page_id: ID,
        background_type_id: ID,
    ) => Promise<true>;
    setPageByIds: (
        book_id: ID,
        page_ids: ID[],
        background_type_id: ID,
    ) => Promise<true>;

    setPageByNumber: (
        book_id: ID,
        page_number: number,
        background_type_id: ID,
    ) => Promise<true>;

    setPageByNumbers: (
        book_id: ID,
        page_numbers: number[],
        background_type_id: ID,
    ) => Promise<true>;

    getSelectedByBookId: (book_id: ID) => Promise<ID>;
}
